from typing import Any, Callable, Iterable, Optional, Union

import itertools
import math

from abc import ABC, abstractmethod

import numpy as np

from mckit.constants import FLOAT_TOLERANCE
from numpy import ndarray


class MaybeClose(ABC):
    @abstractmethod
    def is_close_to(self, other: Any, estimator: "EstimatorType") -> bool:
        """Objects can be estimated as close with some estimator"""


ComparableType = Optional[Union[Iterable, ndarray, float, MaybeClose]]
EstimatorType = Callable[[ComparableType, ComparableType], bool]


def tolerance_estimator(
    rtol: float = FLOAT_TOLERANCE,
    atol: float = FLOAT_TOLERANCE,
    equal_nan: bool = False,
) -> EstimatorType:
    """
    Estimates "closeness of numpy arrays and float scalars using math.isclose and numpy.allclose methods
    """

    def _call(a: ComparableType, b: ComparableType) -> bool:
        if a is b:
            return True
        if a is None:
            return False  # b is not None here
        if b is None:
            return True  # in our use cases absent optional objects (like Transformations) are equal
        if isinstance(a, float) and isinstance(b, float):
            return math.isclose(a, b, rel_tol=rtol, abs_tol=atol)
        elif isinstance(a, ndarray) and isinstance(b, ndarray):
            return np.allclose(a, b, rtol, atol, equal_nan)
        elif issubclass(type(a), Iterable) and issubclass(type(b), Iterable):
            for ai, bi in itertools.zip_longest(a, b):
                if not _call(ai, bi):
                    return False
            return True
        elif isinstance(a, int) and isinstance(b, int):
            return a == b
        elif isinstance(a, MaybeClose) and isinstance(b, MaybeClose):
            return a.is_close_to(b, _call)
        else:
            raise TypeError(
                f"Not implemented for {type(a).__name__} and {type(b).__name__}"
            )

    return _call



# flake8: noqa F401
from .accept import TVisitor, accept, on_unknown_acceptor
from .io import (
    MCNP_ENCODING,
    check_if_all_paths_exist,
    check_if_path_exists,
    get_root_dir,
    make_dir,
    make_dirs,
)
from .misc import (
    MAX_DIGITS,
    are_equal,
    compute_hash,
    deepcopy,
    filter_dict,
    get_decades,
    is_in,
    is_sorted,
    make_hashable,
    mids,
    prettify_float,
    round_array,
    round_scalar,
    significant_array,
    significant_digits,
)
from .resource import filename_resolver, path_resolver
from .tolerance import FLOAT_TOLERANCE



from typing import Generator

import os

from pathlib import Path

MCNP_ENCODING = "Cp1251"
"""The encoding used in SuperMC when creating MCNP models code. Some symbols are not Unicode."""


def get_root_dir(environment_variable_name, default):
    return Path(os.getenv(environment_variable_name, default)).expanduser()


def make_dir(d: Path) -> Path:
    d.mkdir(parents=True, exist_ok=True)
    return d


def make_dirs(*dirs: Path) -> Generator[Path, None, None]:
    yield from map(make_dir, dirs)


def check_if_path_exists(p: Path) -> Path:
    if p.exists():
        return p
    raise FileNotFoundError(f'Path "{p}" does not exist')


def check_if_all_paths_exist(*paths: Path) -> Generator[Path, None, None]:
    yield from map(check_if_path_exists, paths)


class FindFileInDirectoriesError(EnvironmentError):
    def __init__(self, _file, directories):
        super().__init__(f"Cannot find {_file} in directories {directories}")
        self.file = _file
        self.directories = directories


def find_file_in_directories(_file: str, *directories: Path) -> Path:
    """Find a file in directories

    Args:
        _file: a file to find
        directories: list of directories to search the file in

    Raises:
        FindFileInDirectoriesError: if the `_file` is not found in the specified `directories`
    """
    for d in directories:
        path = d / _file
        if path.exists():
            return path.absolute()
    raise FindFileInDirectoriesError(_file, directories)



from typing import Callable, Dict, Iterable, NoReturn, Optional, Type, TypeVar, cast

from functools import reduce

from mckit.utils.named import Name, default_name_key

Key = TypeVar("Key")
Item = TypeVar("Item")


class Index(Dict[Key, Item]):
    """Like collections.defaultdict but the factory takes key as argument."""

    def __init__(
        self, default_factory: Callable[[Key], Optional[Item]] = None, **kwargs
    ):
        super().__init__(self, **kwargs)
        self._default_factory = default_factory

    def __missing__(self, key: Key) -> Optional[Item]:
        """
        Calls default factory with the key as argument.
        """
        if self._default_factory:
            df = self._default_factory
            return df(key)

        raise NumberedItemNotFoundError(key)

    @property
    def default_factory(self):
        return self._default_factory


# noinspection PyUnusedLocal
def ignore(item: Key) -> Optional[Item]:
    return None


class IgnoringIndex(Index[Key, Item]):
    def __init__(self) -> None:
        Index.__init__(self, ignore)


class NumberedItemNotFoundError(KeyError):
    kind: str = ""

    def __init__(self, item: Key, *args, **kwargs) -> None:
        super().__init__(args, kwargs)
        self._item = item

    def __str__(self) -> str:
        return f"{self.kind} #{self._item} is not found"


class NumberedItemDuplicateError(ValueError):
    kind: str = ""

    def __init__(self, item: Key, prev: Item, curr: Item, *args, **kwargs) -> None:
        super().__init__(args, kwargs)
        self._item = item
        self._prev = prev
        self._curr = curr

    def __str__(self):
        return f"{self.kind} #{self._item} is duplicated, see {self._prev} and {self._curr}"


def raise_on_duplicate_strategy(key: Key, prev: Item, curr: Item) -> NoReturn:
    raise NumberedItemDuplicateError(key, prev, curr)


def ignore_equal_objects_strategy(key: Key, prev: Item, curr: Item) -> None:
    if prev is not curr and prev != curr:
        raise NumberedItemDuplicateError(key, prev, curr)


class StatisticsCollector(Dict[Key, int]):
    def __init__(self, ignore_equal=False):
        super().__init__(self)
        self.ignore_equal = ignore_equal

    def __missing__(self, key):
        return 1

    def __call__(self, key: Key, prev: Item, curr: Item) -> None:
        if not self.ignore_equal or prev != curr:
            self[key] += 1


class IndexOfNamed(Index[Key, Item]):
    @classmethod
    def from_iterable(
        cls: Type["IndexOfNamed[Key, Item]"],
        items: Iterable[Item],
        *,
        key: Callable[[Item], Name] = default_name_key,
        default_factory: Callable[[Key], Item] = None,
        on_duplicate: Callable[[Key, Item, Item], None] = None,
        **kwargs,
    ) -> "IndexOfNamed[Key, Item]":
        index = cls(default_factory, **kwargs)

        def reducer(a, b):
            name = key(b)
            if on_duplicate is not None:
                if name in a:
                    on_duplicate(name, a[name], b)
            a[name] = b
            return a

        return cast("IndexOfNamed[Key, Item]", reduce(reducer, items, index))
